==> coiled-0.8.5-py3-none-any.whl: coiled/run.py <==
from __future__ import annotations

import functools
import threading
from typing import Optional, Union

import dask.distributed
from dask.base import tokenize

import coiled

_clients = {}
_lock = threading.RLock()
_clusters = {}


class Function:
    """A function that you can run remotely"""

    def __init__(self, function, cluster_kwargs):
        self.function = function
        self._cluster_kwargs = cluster_kwargs

    @functools.cached_property
    def _name(self):
        token = tokenize(**self._cluster_kwargs)
        return f"run-{token[:8]}"

    @property
    def cluster(self) -> coiled.Cluster:
        with _lock:
            try:
                return _clusters[self._name]
            except KeyError:
                cluster = coiled.Cluster(name=self._name, **self._cluster_kwargs)
                cluster.adapt(minimum=0, maximum=100)
                _clusters[self._name] = cluster
                return cluster

    @property
    def client(self) -> dask.distributed.Client:
        with _lock:
            try:
                return _clients[self._name]
            except KeyError:
                client = self.cluster.get_client()
                if self.cluster.shutdown_on_close:
                    client.register_scheduler_plugin(NoClientShutdown())
                _clients[self._name] = client
                return client

    def __call__(self, *args, **kwargs):
        # TODO, priority and stuff
        return self.client.run_on_scheduler(self.function, *args, **kwargs)

    def submit(self, *args, **kwargs) -> dask.distributed.Future:
        """Submit function call for asynchronous execution

        This immediately returns a Dask Future, allowing for the submission of
        many tasks in parallel.

        Example
        -------
        >>> @coiled.run()
        ... def f(x):
        ...    return x + 1

        >>> f(10)  # calling the function blocks until finished
        11
        >>> f.submit(10)  # immediately returns a future
        <Future: pending, key=f-1234>
        >>> f.submit(10).result()  # Call .result to get result
        11

        >>> futures = [f(i) for i in range(1000)]  # parallelize with a for loop
        >>> [future.result() for future in futures]
        ...

        Returns
        -------
        future: dask.distributed.Future
        """
        return self.client.submit(self.function, *args, **kwargs)


def run(
    *,
    software: Optional[str] = None,
    container: Optional[str] = None,
    vm_type: Optional[list[str]] = None,
    cpu: Optional[Union[int, list[int]]] = None,
    memory: Optional[Union[str, list[str]]] = None,
    gpu: Optional[bool] = None,
    account: Optional[str] = None,
    region: Optional[str] = None,
    arm: Optional[bool] = None,
    shutdown_on_close: bool = True,
):
    """
    Decorate a function to run on cloud infrastructure

    This creates a ``Function`` object that executes its code on a remote cluster
    with the hardware and software specified in the arguments to the decorator.
    It can run either as a normal function, or it can return Dask Futures for
    parallel computing.

    See the :class:`coiled.Cluster` docstring for parameter descriptions.

    .. warning::

        ``@coiled.run`` is an experimental feature and is subject to breaking changes.

    Examples
    --------
    >>> import coiled
    >>> @coiled.run()
    ... def f(x):
    ...    return x + 1

    >>> f(10)  # calling the function blocks until finished
    11
    >>> f.submit(10)  # immediately returns a future
    <Future: pending, key=f-1234>
    >>> f.submit(10).result()  # Call .result to get result
    11

    >>> futures = [f(i) for i in range(1000)]  # parallelize with a for loop
    >>> [future.result() for future in futures]
    ...
    """

    def decorator(func) -> Function:
        environ = None
        if container and "rapidsai" in container:
            environ = {"DISABLE_JUPYTER": "true"}  # needed for "stable" RAPIDS image

        cluster_kwargs = dict(
            account=account,
            n_workers=0,
            scheduler_cpu=cpu,
            scheduler_memory=memory,
            software=software,
            container=container,
            idle_timeout="24 hours",
            scheduler_vm_types=vm_type,
            worker_vm_types=vm_type,
            allow_ssh=True,
            environ=environ,
            scheduler_gpu=gpu,
            region=region,
            arm=arm,
            shutdown_on_close=shutdown_on_close,
            tags={"coiled-cluster-type": "run/python"},
        )

        return Function(func, cluster_kwargs)

    return decorator


class NoClientShutdown(dask.distributed.SchedulerPlugin):
    """Shut down a scheduler if all of the clients have gone"""

    name = "NoClientShutdown"

    def remove_client(self, scheduler, client):
        if set(scheduler.clients) == {"fire-and-forget"}:
            scheduler._ongoing_background_tasks.call_soon(scheduler.close)

==> coiled-0.8.6.dev14-py3-none-any.whl: coiled/run.py <==
from __future__ import annotations

import functools
import sys
import threading
from typing import Optional, Union

import dask.distributed
from dask.base import tokenize
from dask.utils import parse_timedelta

import coiled

from .shutdown import NoClientShutdown

_clients = {}
_lock = threading.RLock()
_clusters = {}


class Function:
    """A function that you can run remotely"""

    def __init__(self, function, cluster_kwargs, keepalive):
        self.function = function
        self._cluster_kwargs = cluster_kwargs
        self.keepalive = parse_timedelta(keepalive)

    @functools.cached_property
    def _name(self):
        token = tokenize(
            sys.executable,
            # TODO: include something about the software environment
            **self._cluster_kwargs,
        )
        return f"run-{token[:8]}"

    @property
    def cluster(self) -> coiled.Cluster:
        with _lock:
            try:
                return _clusters[self._name]
            except KeyError:
                cluster = coiled.Cluster(name=self._name, **self._cluster_kwargs)
                cluster.adapt(minimum=0, maximum=100)
                _clusters[self._name] = cluster
                return cluster

    @property
    def client(self) -> dask.distributed.Client:
        with _lock:
            try:
                return _clients[self._name]
            except KeyError:
                client = self.cluster.get_client()
                if self.cluster.shutdown_on_close:
                    client.register_scheduler_plugin(NoClientShutdown(keepalive=self.keepalive))
                _clients[self._name] = client
                return client

    def __call__(self, *args, **kwargs):
        # TODO, priority and stuff
        return self.client.run_on_scheduler(self.function, *args, **kwargs)

    def local(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def submit(self, *args, **kwargs) -> dask.distributed.Future:
        """Submit function call for asynchronous execution

        This immediately returns a Dask Future, allowing for the submission of
        many tasks in parallel.

        Example
        -------
        >>> @coiled.run()
        ... def f(x):
        ...    return x + 1

        >>> f(10)  # calling the function blocks until finished
        11
        >>> f.submit(10)  # immediately returns a future
        <Future: pending, key=f-1234>
        >>> f.submit(10).result()  # Call .result to get result
        11

        >>> futures = [f(i) for i in range(1000)]  # parallelize with a for loop
        >>> [future.result() for future in futures]
        ...

        Returns
        -------
        future: dask.distributed.Future
        """
        return self.client.submit(self.function, *args, **kwargs)


def run(
    *,
    software: Optional[str] = None,
    container: Optional[str] = None,
    vm_type: Optional[list[str]] = None,
    cpu: Optional[Union[int, list[int]]] = None,
    memory: Optional[Union[str, list[str]]] = None,
    gpu: Optional[bool] = None,
    account: Optional[str] = None,
    region: Optional[str] = None,
    arm: Optional[bool] = None,
    shutdown_on_close: bool = True,
    keepalive="30 seconds",
):
    """
    Decorate a function to run on cloud infrastructure

    This creates a ``Function`` object that executes its code on a remote cluster
    with the hardware and software specified in the arguments to the decorator.
    It can run either as a normal function, or it can return Dask Futures for
    parallel computing.

    See the :class:`coiled.Cluster` docstring for parameter descriptions.

    .. warning::

        ``@coiled.run`` is an experimental feature and is subject to breaking changes.

    Examples
    --------
    >>> import coiled
    >>> @coiled.run()
    ... def f(x):
    ...    return x + 1

    >>> f(10)  # calling the function blocks until finished
    11
    >>> f.submit(10)  # immediately returns a future
    <Future: pending, key=f-1234>
    >>> f.submit(10).result()  # Call .result to get result
    11

    >>> futures = [f(i) for i in range(1000)]  # parallelize with a for loop
    >>> [future.result() for future in futures]
    ...
    """

    def decorator(func) -> Function:
        environ = None
        if container and "rapidsai" in container:
            environ = {"DISABLE_JUPYTER": "true"}  # needed for "stable" RAPIDS image

        cluster_kwargs = dict(
            account=account,
            n_workers=0,
            scheduler_cpu=cpu,
            scheduler_memory=memory,
            software=software,
            container=container,
            idle_timeout="24 hours",
            scheduler_vm_types=vm_type,
            worker_vm_types=vm_type,
            allow_ssh=True,
            environ=environ,
            scheduler_gpu=gpu,
            region=region,
            arm=arm,
            shutdown_on_close=shutdown_on_close,
            tags={"coiled-cluster-type": "run/python"},
        )

        return Function(func, cluster_kwargs, keepalive=keepalive)

    return decorator
